#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string>
#include "Logger.hpp"
#include "AgentCommInterface.hpp"
#include "utility.hpp"
#include "PredParam.hpp"
#include "PPSharedParam.hpp"
#include "PredatorTagFunction.hpp"
#include "PPSharedTypes.hpp"
#include "random.hpp"
#include "RandomBehavior.hpp"
#include "FixedPolicyBehavior.hpp"
#include "SimpleQLearnBehavior.hpp"
#include "StrictCoachReqBehavior.hpp"
#include "QLearnWCoachBehavior.hpp"
#include "SimpleQLearnCSABehavior.hpp"
#include "FIFOLimSAAdviceRepository.hpp"

using namespace spades;
using namespace spadesalib;

/*****************************************************************************************/

//this is static so that the signal handlers can access it
static spadesalib::AgentCommInterface* pacs_interface;
  

void init_handler_term();
void init_handler_segv();

/*****************************************************************************************/

class ACSITimeTagger : public TimeTagger
{
public:
  ACSITimeTagger(spadesalib::AgentCommInterface* p) : pacs(p) {}
  std::string getTimeTag() { return toString(pacs->getLatestTime()); }

private:
  spadesalib::AgentCommInterface* pacs;
};


/*****************************************************************************************/

int main (int argc, char** argv)
{
  PredParam* pParam = new PredParam(argc, argv);

  PredatorTagFunction *pTagFunc = new PredatorTagFunction;

  spades::Logger::instance()->setTagFunction(pTagFunc);
  spades::Logger::instance()->setMaxActionLogLevel(pParam->getActionLogLevel());

  init_handler_term();
  init_handler_segv();

  spades::seedRandom(pParam->getRandomSeed());

  pacs_interface = new AgentCommInterface();

  pTagFunc->setTimeTagger(new ACSITimeTagger(pacs_interface));
  
  Behavior* pbeh = NULL;
  switch (pParam->getActMode())
    {
    case AM_None:
      errorlog << "I don't know what to do with act mode none " << pParam->getActMode() << ende;
      break;
    case AM_Random:
      pbeh = new RandomBehavior(pacs_interface);
      break;
    case AM_Policy:
      pbeh = new FixedPolicyBehavior(pacs_interface, pParam->getFixedPolicyFN().c_str());
      break;
    case AM_QLearn: {
      SimpleQLearnBehavior* p = 
	new SimpleQLearnBehavior(pacs_interface,
				 pParam->getGamma(),
				 pParam->getQLearnExplore(),
				 pParam->getQLearnExploreProb(),
				 pParam->getQLearnBoltzmannConstant(),
				 pParam->getQLearnDisableMapFN().c_str());
      p->setFinalPolicyFN(pParam->getQLearnFinalPolicyFN());
      p->setFinalQTableFN(pParam->getQLearnFinalQTableFN());
      p->setPolicyChangeFN(pParam->getQLearnPolicyChangeFN());
      p->setPolicyEval(pParam->getQLearnPolicyEvalFN(),
		       pParam->getQLearnPolicyEvalInterval(),
		       pParam->getQLearnPolicyEvalLength());
      pbeh = p;
    } break;
    case AM_StrictCoachReq:
      pbeh = new StrictCoachReqBehavior(pacs_interface);
      break;
    case AM_QLearnWCoach: {
      QLearnWCoachBehavior* p =
	 new QLearnWCoachBehavior(pacs_interface,
				  pParam->getGamma(),
				  pParam->getQLearnExplore(),
				  pParam->getQLearnExploreProb(),
				  pParam->getQLearnBoltzmannConstant(),
				  pParam->getQLearnWCoachCoachProb(),
				  pParam->getQLearnWCoachRandomOnAdvisedDisable(),
				  pParam->getQLearnDisableMapFN().c_str());
      p->setFinalPolicyFN(pParam->getQLearnFinalPolicyFN());
      p->setFinalQTableFN(pParam->getQLearnFinalQTableFN());
      p->setPolicyChangeFN(pParam->getQLearnPolicyChangeFN());
      p->setPolicyEval(pParam->getQLearnPolicyEvalFN(),
		       pParam->getQLearnPolicyEvalInterval(),
		       pParam->getQLearnPolicyEvalLength());
      pbeh = p;
    } break;
    case AM_QLearnCSASimple: {
      SimpleQLearnCSABehavior* p = 
	 new SimpleQLearnCSABehavior(pacs_interface,
				     pParam->getGamma(),
				     pParam->getQLearnExplore(),
				     pParam->getQLearnExploreProb(),
				     pParam->getQLearnBoltzmannConstant(),
				     pParam->getQLearnCSACoachProb(),
				     pParam->getQLearnDisableMapFN().c_str());      
      switch (PredParam::instance()->getSAARType()) {
      case SAAR_None:
	errorlog << "I don't know wha to do with SAAR_None" << ende;
	break;
      case SAAR_Complete:
	p->setSAAR(new SAAdviceRepository(State::getNumStates()));
	break;
      case SAAR_FIFO:
	p->setSAAR(new FIFOLimSAAdviceRepository(State::getNumStates(), PredParam::instance()->getSAARFIFOLimit()));
	break;
      default:
	errorlog << "I don't understand saar_type: " << PredParam::instance()->getSAARType() << ende;
	break;
      }
      p->setFinalPolicyFN(pParam->getQLearnFinalPolicyFN());
      p->setFinalQTableFN(pParam->getQLearnFinalQTableFN());
      p->setFinalAdviceFN(pParam->getSAAdviceFinalAdviceFN());
      p->setPolicyChangeFN(pParam->getQLearnPolicyChangeFN());
      p->setPolicyEval(pParam->getQLearnPolicyEvalFN(),
		       pParam->getQLearnPolicyEvalInterval(),
		       pParam->getQLearnPolicyEvalLength());
      pbeh = p;
    } break;
    default:
      errorlog << "What is act mode? " << pParam->getActMode() << ende;
      break;
    }
  
  actionlog(10) << "Predator started" << ende;

  int res = pbeh->behaviorLoop();
  //int res = mainLoop(pacs_interface);

  actionlog(10) << "Client is exiting after " << res << " messages"
		<< ende;

  delete pbeh;
  pbeh = NULL;

  Logger::removeInstance();

  delete pParam;
  pParam = NULL;

  return res;
  
}


/*****************************************************************************************/

void sigterm_handler(int) 
{
  pacs_interface->initiateShutdown();
}

void init_handler_term() 
{
  struct sigaction sigact;
  sigact.sa_flags = 0;

  //This makes sure that the handler is only called once for each signal
  sigact.sa_flags |= SA_RESETHAND;
  
  sigemptyset(&sigact.sa_mask);
  sigact.sa_handler = sigterm_handler; 
  sigaction(SIGINT, &sigact, NULL);
  sigaction(SIGQUIT, &sigact, NULL);
  sigaction(SIGTERM, &sigact, NULL);
}

void sigsegv_handler(int) 
{
  errorlog << "Got a seg fault!" << ende;
  pacs_interface->initiateShutdown();
}

void init_handler_segv() 
{
  struct sigaction sigact;
  sigact.sa_flags = 0;

  //This makes sure that the handler is only called once for each signal
  sigact.sa_flags |= SA_RESETHAND;
  
  sigemptyset(&sigact.sa_mask);
  sigact.sa_handler = sigsegv_handler; 
  sigaction(SIGSEGV, &sigact, NULL);
}
