/* This is a virtual base class behavior for qlearn style behaviors
   it sets up some logging and other recording, as well as storing
   actions and such */

#include "BaseQLearnBehavior.hpp"
#include "Logger.hpp"
#include "State.hpp"
#include "Policy.hpp"
#include "PPSharedParam.hpp"

using namespace std;
using namespace spades;
using namespace spadesalib;


BaseQLearnBehavior::BaseQLearnBehavior(AgentCommInterface* pacs,
				       float gamma,
				       explore_t exp, float alpha_explore, float boltzmann_constant,
				       const char* disable_map_fn)
  : Behavior(pacs),
    qtable(State::getNumStates(), NUM_BASIC_MOVE_T, gamma),
    exp_type(exp), alpha_explore(alpha_explore), boltzmann_constant(boltzmann_constant),
    disable_map(disable_map_fn),
    last_valid(false), last_state(), last_action(BM_None),
    policy_fn(), qtable_fn(),
    os_policy_change(), policy_change_count(0),
    last_policy_change_time(0), stat_policy_time_limit(-1),
    os_policy_eval(), policy_eval_interval(-1), policy_eval_length(-1),
    policy_eval_current(false), policy_eval_last_time(0), policy_eval_last_rew_time(0), policy_eval_mean(),
    actmap_compare_interval(-1), actmap_compare_pactmap(NULL), actmap_compare_os(), actmap_compare_last_time(-1),
    actmap_action_compare_interval(-1), actmap_action_compare_pactmap(NULL), actmap_action_compare_os(), actmap_action_compare_last_time(0), actmap_action_compare_summ(),
    forbid_q_update_next(false)
{
  qtable.disableSet(disable_map);
}

BaseQLearnBehavior::BaseQLearnBehavior(AgentCommInterface* pacs,
				       const char* init_qtable_fn,
				       explore_t exp, float alpha_explore, float boltzmann_constant,
				       const char* disable_map_fn)
  : Behavior(pacs),
    qtable(init_qtable_fn),
    exp_type(exp), alpha_explore(alpha_explore), boltzmann_constant(boltzmann_constant),
    disable_map(disable_map_fn),
    last_valid(false), last_state(), last_action(BM_None),
    policy_fn(), qtable_fn(),
    os_policy_change(), policy_change_count(0),
    last_policy_change_time(0), stat_policy_time_limit(-1),
    os_policy_eval(), policy_eval_interval(-1), policy_eval_length(-1),
    policy_eval_current(false), policy_eval_last_time(0), policy_eval_last_rew_time(0), policy_eval_mean(),
    forbid_q_update_next(false)
{
  qtable.disableSet(disable_map);
}

BaseQLearnBehavior::~BaseQLearnBehavior()
{
  if (!policy_fn.empty())
    {
      Policy p(State::getNumStates());
      p.extractFromQTable(&qtable);

      ofstream outfile(policy_fn.c_str());
      if (!outfile)
	{
	  errorlog << "BaseQLearnBehavior: could not open policy output file '" << policy_fn << "'" << ende;
	}
      else
	{
	  outfile << p;
	  outfile.close();
	}
    }

  if (!qtable_fn.empty())
    {
      ofstream outfile(qtable_fn.c_str());
      if (!outfile)
	{
	  errorlog << "BaseQLearnBehavior: could not open qtable output file '" << qtable_fn << "'" << ende;
	}
      else
	{
	  outfile << qtable;
	  outfile.close();
	}
    }
  
  writePolicyChangeCount();

  if (actmap_compare_pactmap)
    delete actmap_compare_pactmap;
  
  if (actmap_action_compare_pactmap)
    delete actmap_action_compare_pactmap;
  
}

bool
BaseQLearnBehavior::initialize()
{
  return true;
}


/***************************************************************************************/

void
BaseQLearnBehavior::setPolicyChangeFN(std::string fn)
{
  if (fn.empty())
    {
      os_policy_change.close();
      return;
    }
  
  os_policy_change.open(fn.c_str());
  if (!os_policy_change)
    {
      errorlog << "Could not open policy change file: " << fn << ende;
      return;
    }
  
  os_policy_change << "# This file was created by BaseQLearnBehavior" << endl
		   << "# Format: <time> <number of policy changes>" << endl;
}


void
BaseQLearnBehavior::setPolicyEval(std::string fn, int interval, int length)
{
  os_policy_eval.open(fn.c_str());
  if (!os_policy_eval)
    errorlog << "Could not open policy eval file '" << fn << "'" << ende;
  policy_eval_interval = interval;
  policy_eval_length = length;
}

void
BaseQLearnBehavior::setCompareToActMap(int interval, ActionSubsetMap* pActMap, std::string fn)
{
  if (actmap_compare_pactmap)
    delete actmap_compare_pactmap;

  actmap_compare_interval = interval;
  if (interval <= 0)
    return;

  actmap_compare_pactmap = pActMap;

  actmap_compare_os.open(fn.c_str());
  if (!actmap_compare_os)
    errorlog << "Could not open actmap compare file '" << fn << "'" << ende;

  actmap_compare_os << "# Generated by the actmap_compare feature fo BaseQLearnBehavior" << endl;
  actmap_compare_os << "# Format: <time> <incnt> <outcnt> <skipcnt>" << endl;
}

// compares the actions taken to those in an act map
// note that you must explictly say when you take an action! This is different
// from the setLastAction function, which deals only with the q-update.
// if interval is <= 0, all other params ignored and this is disabled
void
BaseQLearnBehavior::setActionCompareToActMap(int interval, ActionSubsetMap* pActMap, std::string fn)
{
  if (actmap_action_compare_pactmap)
    delete actmap_action_compare_pactmap;

  actmap_action_compare_interval = interval;
  if (interval <= 0)
    return;

  actmap_action_compare_pactmap = pActMap;

  actmap_action_compare_os.open(fn.c_str());
  if (!actmap_action_compare_os)
    errorlog << "Could not open actmap action_compare file '" << fn << "'" << ende;

  actmap_action_compare_os << "# Generated by the actmap_action_compare feature fo BaseQLearnBehavior" << endl;
  actmap_action_compare_os << "# Format: <time> <num points> <% in actmap>" << endl;
}

/***************************************************************************************/

void
BaseQLearnBehavior::respondSense(SenseType st)
{
  SimTime curr_time = pacs->getLatestTime();

  handleActmapCompare();
  handleActmapCompareAction();
  
  if (policy_eval_current)
    {
      if (st == ST_State)
	{
	  actionlog(30) << "Policy eval: processing reward: " << getLastReward() << ende;

	  policy_eval_mean.addPoint((float)getLastReward() / (float)(curr_time - policy_eval_last_rew_time),
				    curr_time - policy_eval_last_rew_time);
	  policy_eval_last_rew_time = curr_time;
	  
	  if (curr_time - policy_eval_last_time > policy_eval_length)
	    {
	      actionlog(30) << "Policy eval: finishing eval epoch" << ende;
	      os_policy_eval << curr_time << '\t' << policy_eval_mean.getMean() << endl;
	      policy_eval_current = false;
	      policy_eval_last_time = curr_time;
	      policy_eval_mean.reset();
	      //this gives the sub class the chance to respond to this
	      respondSenseQLearn(st);
	    }
	  else
	    {
	      respondForPolicyEval();
	    }
	}
      else
	{
	  actionlog(30) << "Ignoring sense " << st << " while policy eval" << ende;
	  /* it's not entirely clear to me if this is the right thing to do here
	     The agent could send out actions here which override our already sent move,
	     which would be bad. The send functions are al defined in Behavior, so there's
	     really no clean way I can see to print out errors if that occurs.
	     I think I'll just have to be careful to initiate stuff only on
	     ST_State messages, so if those messages are not received, then no
	     actions will be sent */
	  respondSenseQLearn(st);
	}
    }
  else
    {
      if (st != ST_State ||
	  policy_eval_interval <= 0 || policy_eval_length <= 0 ||
	  curr_time - policy_eval_last_time <= policy_eval_interval)
	{
	  respondSenseQLearn(st);
	}
      else
	{
	  actionlog(30) << "Policy eval: beginning" << ende;
	  policy_eval_current = true;
	  policy_eval_last_time = curr_time;
	  policy_eval_last_rew_time = curr_time;
	  respondForPolicyEval();
	}
    }
}


void
BaseQLearnBehavior::respondForPolicyEval()
{
  sendMove((basic_move_t)qtable.getActMaxRandom(getCurrState().flatVal()));
}

void
BaseQLearnBehavior::handleActmapCompare()
{
  if (actmap_compare_interval <= 0)
    return;

  if (actmap_compare_pactmap == NULL)
    errorlog << "actmap_compare_pactmap should not be NULL" << ende;

  int curr_time = pacs->getLatestTime();
  
  if (actmap_compare_last_time == -1 ||
      curr_time - actmap_compare_last_time >= actmap_compare_interval)
    {
      actionlog(30) << "Starting a comparision to the given action map" << ende;
  
      Policy p(actmap_compare_pactmap->getNumStates());
      CaptureStateFilter filter(false);
      int cnt_in, cnt_out, cnt_skip;
      p.extractFromQTable(&qtable);

      actmap_compare_pactmap->compareToPolicy(p, &filter, &cnt_in, &cnt_out, &cnt_skip);

      actmap_compare_os << curr_time << '\t'
			<< cnt_in << '\t'
			<< cnt_out << '\t'
			<< cnt_skip << endl;

      actmap_compare_last_time = curr_time;
      
      actionlog(30) << "Comparison done: " << cnt_in << ' ' << cnt_out << ' ' << cnt_skip << ende;
    }

}

void
BaseQLearnBehavior::actForActionCompare(unsigned state, basic_move_t m)
{
  if (actmap_action_compare_pactmap == NULL)
    return;
  
  actionlog(150) << "actForActionCompare: adding " << m << ende;
  actmap_action_compare_summ.addPoint(actmap_action_compare_pactmap->isIn(state, (int)m) ? 1.0 : 0.0);
}


void
BaseQLearnBehavior::handleActmapCompareAction()
{
  if (actmap_action_compare_interval <= 0)
    return;

  if (actmap_action_compare_pactmap == NULL)
    errorlog << "actmap_action_compare_pactmap should not be NULL" << ende;

  int curr_time = pacs->getLatestTime();
  
  if (curr_time - actmap_action_compare_last_time >= actmap_action_compare_interval)
    {
      actionlog(150) << "Recording action to actmap comparison to file" << ende;
      
      actmap_action_compare_os << curr_time << '\t'
			       << actmap_action_compare_summ.getNumPoints() << '\t'
			       << actmap_action_compare_summ.getMean() << endl;

      actmap_action_compare_summ.reset();

      actmap_action_compare_last_time = curr_time;
    }
}



/***************************************************************************************/

//uses last state, last action, getCurrState(), getCurrReward()
void
BaseQLearnBehavior::updateQValue()
{
  if (!last_valid)
    {
      actionlog(50) << "updateQValue: doing nothing because I have no last state" << ende;
      return;
    }
  
  if (forbid_q_update_next)
    {
      actionlog(50) << "updateQValue: doing nothing because forbid flag is set" << ende;
      return;
    }

  if (PPSharedParam::instance()->getForbidAllQUpdates())
    {
      actionlog(50) << "updateQValue: all q updates forbidden" << ende;
      return;
    }

  actionlog(50) << "updateQValue: Doing a q-upadate for s a s r: "
		<< last_state.flatVal() << ' '
		<< last_action << ' '
		<< getCurrState().flatVal() << ' '
		<< getLastReward() << ende;
  
  if (qtable.observedUpdate(last_state.flatVal(), last_action, getCurrState().flatVal(), getLastReward(),
			    PPSharedParam::instance()->getDeterministicUpdate()))
    {
      actionlog(100) << "Did an update that caused policy change" << ende;
      policy_change_count++;
      writePolicyChangeCount();
      last_policy_change_time = pacs->getLatestTime();
    }
  else
    {
      actionlog(100) << "Did an update that did not cause a policy change" << ende;
      if (stat_policy_time_limit > 0 && 
	  pacs->getLatestTime() - last_policy_change_time > stat_policy_time_limit)
	{
	  actionlog(10) << "It's been a long time since policy changed, so I'm exiting: "
			<< pacs->getLatestTime() - last_policy_change_time
			<< ende;
	  sendShutdown();
	}
      
    }
  
}

/***************************************************************************************/

//this should be called at the end of every State sensation;
//this will update the last state
void
BaseQLearnBehavior::stateSenseUpdate()
{
  last_valid = true;
  last_state = getCurrState();
  if (getVecCapturedAgents().empty())
    {
      forbid_q_update_next = false;
    }
  else
    {
      actionlog(50) << "Forbidding next q update" << ende;
      forbid_q_update_next = true;
    }
}

  
const State&
BaseQLearnBehavior::getLastState() const
{
  if (!last_valid)
    errorlog << "Don't ask for last_state when it's not valid" << ende;
  return last_state;
} 

/***************************************************************************************/

void
BaseQLearnBehavior::writePolicyChangeCount()
{
  os_policy_change << pacs->getLatestTime() << '\t' << policy_change_count << endl;  
}

