/* The main file for a simple text based log player */

#include <curses.h>
#include <stdlib.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <deque>
#include "WorldState.hpp"

using namespace std;

/* valid keys:
   left/right: signle step advance, works only when stopped
   '0' or ')': skip 1000 steps forward 
   '>' or '.' or ' ' : play forward
   ',' or '<': play backward
   'p' or <ENTER>: stop
   'a' or 'A': speed up
   'z' or 'Z': slow down
   'q' or 'Q': quit */

const unsigned cycles_stored = 100;

enum PlayState{
  PS_Play,
  PS_PlayBack,
  PS_Stop,
  PS_Skip
};


int main(int argc, char** argv)
{
  //let's try to open up the file.
  if (argc <= 1)
    {
      cerr << "usage: monitor <log_fn>" << endl;
      return -1;
    }

  ifstream logfile(argv[1]);
  if (!logfile)
    {
      cerr << "Could not open file '" << argv[1] << "'" << endl;
      return -1;
    }

  //ncurses initialization
  initscr();
  cbreak();
  noecho();
  nodelay(stdscr, TRUE);
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);

  if (!WorldState::readHeaderInfo(logfile))
    {
      cerr << "Error reading header info" << endl;
      return -1;
    }
  
  //Now we get the main part of the code
  typedef deque<WorldState> deque_ws_t;
  deque_ws_t deq_ws;
  deque_ws_t::iterator itercurr = deq_ws.end();
  PlayState ps = PS_Play;
  unsigned long sleeptime = 500000;
  bool exit = false;
  bool draw = true;
  int skip_count = 0;

  while (!exit)
    {
      WorldState ws;

      //process any key presses
      int ch;
      bool done = false;
      while ( !done && (ch = getch()) != ERR )
	{
	  switch (ch)
	    {
	    case KEY_LEFT:
	      if (ps == PS_Stop)
		{
		  if (itercurr != deq_ws.begin()) itercurr--;
		  done = true;
		}
	      break;
	    case KEY_RIGHT:
	      if (ps == PS_Stop)
		{
		  if (itercurr != deq_ws.end()) itercurr++;
		  done = true;
		}
	      break;
	    case '0':
	    case ')':
	      ps = PS_Skip;
	      draw = false;
	      skip_count = 1000;
	      break;
	    case '>':
	    case '.':
	    case ' ':
	      ps = PS_Play;
	      break;
	    case ',':
	    case '<':
	      ps = PS_PlayBack;
	      break;
	    case 'p':
	    case KEY_ENTER:
	      ps = PS_Stop;
	      break;
	    case 'a':
	    case 'A':
	      sleeptime -= 10000;
	      break;
	    case 'z':
	    case 'Z':
	      sleeptime += 10000;
	      break;
	    case 'q':
	    case 'Q':
	      exit = true;
	      break;
	    }
	}
  
      //figure out what cycle we want
      switch (ps)
	{
	case PS_Play:
	  if (itercurr != deq_ws.end()) itercurr++;
	  break;
	case PS_PlayBack:
	  if (itercurr != deq_ws.begin()) itercurr--;
	  break;
	case PS_Stop:
	  break;
	case PS_Skip:
	  if (itercurr != deq_ws.end()) itercurr++;
	  if (--skip_count == 0)
	    {
	      ps = PS_Stop;
	      draw = true;
	    }
	  break;
	default:
	  cerr << "What's PlayState? " << ps << endl;
	  break;
	}

      if (itercurr == deq_ws.end())
	{
	  //read a file line
	  logfile >> ws;
	  if (!logfile)
	    break;
	  deq_ws.push_back(ws);
	  itercurr = deq_ws.end();
	  itercurr--;
	  if (deq_ws.size() > cycles_stored)
	    deq_ws.pop_front();
	}

      if (draw)
	{
	  //draw it on the screen
	  itercurr->drawOnNCurses();

	  refresh();
	  usleep(sleeptime);
	}
    }
  
  //ncurses finalization
  endwin();
  
}

#ifdef OLD_TEST_CODE

  //Some simple test code
  if (mvaddch(2, 0, 'a') == ERR)
    exit(-1);
  refresh();
  sleep(1);
  if (mvaddch(2, 10, 'b') == ERR)
    exit(-1);
  refresh();
  sleep(1);
  if (mvaddch(1, 5, 'c') == ERR)
    exit(-1);
  refresh();
  sleep(2);
  
#endif
